import re
from typing import Any, Dict, List, Optional, Union

from returns.pipeline import flow
from returns.pointfree import bind
from returns.result import Failure, ResultE, Success, safe

from piri.casting import get_casting_function
from piri.constants import (  # noqa: WPS235
    CONDITION,
    CONTAINS,
    DEFAULT_GROUP,
    FROM,
    GROUP,
    IN,
    IS,
    NOT,
    ORIGINAL_FORMAT,
    OTHERWISE,
    SEARCH,
    TARGET,
    THEN,
    TO,
)
from piri.valuetypes import MapValue, NewValue, ValueTypes


@safe
def apply_if_statements(if_value, if_objects) -> MapValue:
    """Apply if statements to a value.

    :param if_value: The value to use when evaluating if statements
    :type if_value: MapValue

    :param if_objects: :term:`if_objects` collection of if operations
    :type if_objects: List[Dict[str, Any]]

    :return: Success/Failure containers
    :rtype: MapValue

    one if object looks like this

    .. code-block:: json

        {
            "condition": "is",
            "target": "foo",
            "then": "bar",
            "otherwise": "no foo" - optional
        }

    The if statements are chained so that the next works on the output of the
    previous. If no "otherwise" is provided then the original value or value
    from the previous operation will be returned.

    Example
        >>> apply_if_statements(
        ...     '1', [{'condition': 'is', 'target': '1', 'then': '2'}],
        ... ).unwrap() == '2'
        True
        >>> apply_if_statements(
        ...     'a',
        ...     [{
        ...         'condition': 'is',
        ...         'target': '1',
        ...         'then': '2',
        ...         'otherwise': '3'
        ...     }],
        ... ).unwrap() == '3'
        True

    """
    for if_object in if_objects:

        if_value = _apply_statement(
            if_value, if_object,
        ).value_or(if_value)

    if if_value is None:
        raise ValueError('If statement failed or produced `None`')

    return if_value


@safe
def _apply_statement(
    if_value: Optional[MapValue],
    if_object: Dict[str, Any],
) -> Optional[MapValue]:
    evaluation: bool = False

    condition = if_object[CONDITION]
    target = if_object[TARGET]

    if condition == IS:
        evaluation = if_value == target

    if condition == NOT:
        evaluation = if_value != target

    if condition == IN:
        evaluation = if_value in target

    if condition == CONTAINS:
        list_or_dict = isinstance(if_value, (dict, list))
        evaluation = list_or_dict and target in if_value  # type: ignore
        evaluation = evaluation or not list_or_dict and target in str(if_value)

    if evaluation:
        return if_object[THEN]

    return if_object.get(OTHERWISE, if_value)


@safe
def apply_separator(
    mapped_values: List[MapValue],
    separator: str,
) -> MapValue:
    """Apply separator between the values of a List[Any].

    :param mapped_values: The list of values to join with the separator
    :type mapped_values: List[MapValue]

    :param separator: :term:`separator` value to join mapped_values list with
    :type separator: str

    :return: Success/Failure containers
    :rtype: MapValue

    Example
        >>> from returns.pipeline import is_successful
        >>> apply_separator(['a', 'b', 'c'], ' ').unwrap()
        'a b c'
        >>> apply_separator([1, 'b', True], ' ').unwrap()
        '1 b True'
        >>> is_successful(apply_separator([], ' '))
        False

    """
    if not mapped_values:
        raise ValueError('mapped_values is empty')

    if len(mapped_values) == 1:
        return mapped_values[0]

    return separator.join([str(mapped) for mapped in mapped_values])


def apply_slicing(
    value_to_slice: Optional[NewValue],
    slicing: Dict[str, Any],
) -> Optional[NewValue]:
    """Slice value from index to index.

    :param value_to_slice: The value to slice
    :type value_to_slice: MapValue

    :param slicing: :term:`slicing` object
    :type slicing: dict


    :return: Success/Failure containers
    :rtype: MapValue

    Example
        >>> apply_slicing('123', {'from': 1})
        '23'
        >>> apply_slicing('test', {'from': 1, 'to': 3})
        'es'
    """
    if value_to_slice is None:
        return value_to_slice

    if not slicing:
        return value_to_slice

    if not isinstance(value_to_slice, list):
        value_to_slice = str(value_to_slice)

    return value_to_slice[slicing[FROM]:slicing.get(TO)]


@safe
def apply_regexp(  # noqa: WPS212, WPS234
    value_to_match: Optional[MapValue],
    regexp: Dict[str, Any],
) -> Union[List[MapValue], MapValue, None]:
    r"""Match value by a certain regexp pattern.

    :param value_to_match: The value to match
    :type value_to_match: MapValue

    :param regexp: :term: `matching` object which has parameters for
        regexp match
    :type regexp: dict

    :return: Success/Failure container
    :rtype: MapValue

    Example
        >>> from piri.functions import apply_regexp
        >>> apply_regexp('abcdef', {'search': '(?<=abc)def'}).unwrap()
        'def'
        >>> apply_regexp(
        ...     'Isaac Newton, physicist',
        ...     {'search': r'(\w+)', 'group': 1},
        ... ).unwrap()
        'Newton'
        >>> apply_regexp(
        ...     'r7/p4pN1/1pn4k/8/2bP3R/2P3R1/6PP/6K1 b - -',
        ...     {'search': r'(P[\d|\w])', 'group': [0, 2]},
        ... ).unwrap()
        ['P3', 'PP']
        >>> apply_regexp(None, {'search': 'a+'}).unwrap()
        >>> apply_regexp('lichess rocks!', {'search': None}).unwrap()
        'lichess rocks!'
        >>> apply_regexp('Open-source matters', None).unwrap()
        'Open-source matters'
        >>> apply_regexp(
        ...     '1. e2 e4 d2 d4 2. Nc3 Nc6 3. Qe2 Qe7',
        ...     {'search': r'(e\d)', 'group': []},
        ... ).unwrap()
        ['e2', 'e4', 'e2', 'e7']
        >>> apply_regexp(
        ...     '[Event \"Live Chess\"]\n[Site \"Chess.com\"]\n[Date ',
        ...     {'search': r'Event \"[\d\w ]+\"'}
        ... ).unwrap()
        'Event "Live Chess"'
    """
    if value_to_match is None:
        return value_to_match

    if not regexp or not regexp[SEARCH]:
        return value_to_match

    pattern = regexp[SEARCH]
    groups = re.finditer(pattern, value_to_match)
    matches: list = [gr.group(0) for gr in groups]
    num_group: Union[int, list] = regexp.get(GROUP, DEFAULT_GROUP)
    if isinstance(num_group, list):
        if not num_group:
            return matches
        return [matches[ind] for ind in num_group]  # typing: ignore
    return matches[num_group]


def apply_casting(
    value_to_cast: Optional[MapValue],
    casting: Dict[str, Any],
) -> ResultE[MapValue]:
    """Cast one type of code to another.

    :param casting: :term:`casting` object
    :type casting: dict

    :param value_to_cast: The value to cast to casting['to']
    :type value_to_cast: MapValue

    :return: Success/Failure containers
    :rtype: MapValue

    Example
        >>> apply_casting('123', {'to': 'integer'}).unwrap()
        123
        >>> apply_casting('123.12', {'to': 'decimal'}).unwrap()
        Decimal('123.12')
    """
    if value_to_cast is None:
        return Failure(ValueError('value_to_cast is empty'))

    if TO not in casting or casting[TO] is None:
        return Success(value_to_cast)

    return flow(
        casting[TO],
        get_casting_function,
        bind(  # type: ignore
            lambda function: function(  # type: ignore
                value_to_cast, casting.get(ORIGINAL_FORMAT),
            ),
        ),
    )


@safe
def apply_default(
    mapped_value: Optional[Any] = None,
    default: Optional[MapValue] = None,
) -> MapValue:
    """Apply default value if exists and if mapped value is None.

    :param mapped_value: If this value is *None* the default value is returned
    :type mapped_value: Optional[MapValue]

    :param default: :term:`default` value to return if mapped value is None
    :type default: Optional[MapValue]

    :return: Success/Failure containers
    :rtype: MapValue

    If default *is not* none and mapped_value *is* None then return default
    else if mapped_value is not in accepted ValueTypes then throw an error
    else return mapped value

    Example
        >>> apply_default('test', None).unwrap() == 'test'
        True
        >>> apply_default('nope', 'test').unwrap() == 'nope'
        True
    """
    if default is not None:
        if mapped_value is None:
            return default

    if mapped_value is None:
        raise ValueError('Default value should not be `None`')

    if not isinstance(mapped_value, ValueTypes):
        raise ValueError('Unable to give default value')

    return mapped_value



import decimal
from typing import Any, Dict, List, Optional, Union

from returns.curry import partial
from returns.maybe import maybe
from returns.pipeline import is_successful
from returns.result import safe

from piri.collection_handlers import iterable_data_handler
from piri.constants import (
    ARRAY,
    ATTRIBUTES,
    BRANCHING_ATTRIBUTES,
    BRANCHING_OBJECTS,
    ITERABLES,
    NAME,
    OBJECTS,
    SHOW_EMPTY,
)
from piri.handlers import handle_attribute

decimal.getcontext().rounding = decimal.ROUND_HALF_UP

MappedDict = Dict[str, Any]


@safe
def map_data(
    input_data,
    configuration,
) -> Union[list, dict]:
    """Map entrypoint.

    Try to get iterable data
    if that fails then just run map_object normally, but make it into an array
    if array is true.

    If we had iterable data, iterate that data and run map_object with the
    current iteration data added to the root of the input_data dictionary
    """
    iterate_data = iterable_data_handler(
        input_data, configuration.get(ITERABLES, []),
    )

    if not is_successful(iterate_data):
        return map_object(input_data, configuration).map(
            partial(set_array, array=configuration[ARRAY]),
        ).unwrap()

    mapped_objects: List[dict] = []

    # find returns function to work with iterables
    for iteration in iterate_data.unwrap():
        map_object(
            iteration,
            configuration,
        ).map(
            mapped_objects.append,
        )

    return mapped_objects


def set_array(input_data, array):
    """Return data wrapped in array if if array=True."""
    if array:
        return [input_data]
    return input_data


@maybe
def map_object(input_data, configuration) -> Optional[MappedDict]:
    """Map one object.

    One object has a collections of:
    Attribute mappings,
    Nested object mappings,
    Branching object mappings,

    All functions we call return a dictionary with the mapped values
    so all we have to do is to call update on a shared object_data dict.

    return example:
    return {
        'attrib': 'val',
        'object1': {'attrib1': 'val'}
        'branching_object1: [{'attrib1': 'val'}]
    }
    """
    object_data: MappedDict = {}

    map_attributes(
        input_data, configuration.get(ATTRIBUTES, []),
    ).map(object_data.update)

    map_objects(
        input_data, configuration.get(OBJECTS, []),
    ).map(object_data.update)

    map_branching_objects(
        input_data, configuration.get(BRANCHING_OBJECTS, []),
    ).map(object_data.update)

    # need this as long as empty dict is not seen as None by returns.maybe
    if not object_data and configuration.get(SHOW_EMPTY):
        return {}

    return object_data or None


@maybe
def map_attributes(input_data, configuration) -> Optional[MappedDict]:
    """For all attributes map attribute.

    name of attribute should be set
    {
        'attribute1': 'value',
        'attribute2': 'value2',
    }
    """
    attributes: MappedDict = {}

    for attribute_cfg in configuration:
        attribute_value = handle_attribute(input_data, attribute_cfg)

        if is_successful(attribute_value):
            attributes[attribute_cfg[NAME]] = attribute_value.unwrap()

    return attributes or None


@maybe
def map_objects(
    input_data,
    configuration,
) -> Optional[MappedDict]:
    """For all objects map object.

    name of object should be set.
    {
        'name1': object,
        'name2': object2,
    }
    """
    mapped_objects: MappedDict = {}

    for object_cfg in configuration:
        object_value = map_data(input_data, object_cfg)

        if is_successful(object_value):
            mapped_objects[object_cfg[NAME]] = object_value.unwrap()

    return mapped_objects or None


@maybe
def map_branching_attributes(
    input_data,
    b_attributes,
) -> Optional[List[MappedDict]]:
    """Map branching attributes.

    Branching attributes are a list of attribute mappings that will be
    mapped to the same name in branching object.
    """
    mapped_attributes: List[MappedDict] = []

    for sub_cfg in b_attributes:
        map_attributes(
            input_data, sub_cfg,
        ).map(
            mapped_attributes.append,
        )

    # need this as long as empty dict is not seen as None by returns.maybe
    return mapped_attributes or None


@maybe
def map_branching_objects(
    input_data,
    configuration,
) -> Optional[MappedDict]:
    """Map branching object.

    Branching object is a case where we want to create the same object multiple
    times, however we want to find the data in different places.
    """
    mapped_objects: MappedDict = {}

    for b_object in configuration:
        mapped = map_branching_attributes(
            input_data, b_object[BRANCHING_ATTRIBUTES],
        )

        if is_successful(mapped):
            mapped_objects[b_object[NAME]] = mapped.unwrap()

    # need this as long as empty dict is not seen as None by returns.maybe
    return mapped_objects or None



import json
import os

from attr import dataclass
from jsonschema import Draft7Validator, draft7_format_checker
from returns.result import Failure, ResultE, Success, safe
from typing_extensions import Final, final

from piri.common import ReadLocalFile

SCHEMA: Final[dict] = ReadLocalFile()(
    '{0}/schema.json'.format(os.path.dirname(__file__)), 'r',
).bind(safe(json.loads)).unwrap()


@final
@dataclass(frozen=True, slots=True)
class SchemaValidator(object):
    """Validates data with given JsonSchema Validator."""

    validator: Draft7Validator = Draft7Validator(
        SCHEMA,
        format_checker=draft7_format_checker,
    )

    def __call__(
        self,
        input_data: dict,
    ) -> ResultE[dict]:
        """If is valid return success(input_data) otherwise list of errors."""
        if self.validator.is_valid(input_data):
            return Success(input_data)

        return Failure(
            ValueError(
                list(self.validator.iter_errors(input_data)),
            ),
        )



from typing_extensions import Final

# Objects
NAME = 'name'
ARRAY: Final = 'array'
ITERABLES: Final = 'iterables'

ATTRIBUTES: Final = 'attributes'
OBJECTS: Final = 'objects'
BRANCHING_OBJECTS: Final = 'branching_objects'
BRANCHING_ATTRIBUTES: Final = 'branching_attributes'
SHOW_EMPTY: Final = 'show_empty'

MAPPINGS: Final = 'mappings'
SEPARATOR: Final = 'separator'
IF_STATEMENTS: Final = 'if_statements'
CASTING: Final = 'casting'
DEFAULT: Final = 'default'

# Iterables
ALIAS: Final = 'alias'
PATH: Final = 'path'

# Mapping
# 'path' from iterables
# 'default'

# Regexp
REGEXP: Final[str] = 'regexp'
SEARCH: Final[str] = 'search'
GROUP: Final[str] = 'group'
DEFAULT_GROUP: Final[int] = 0

# IF STATEMENT
CONDITION: Final = 'condition'
IS: Final = 'is'
IN: Final = 'in'
NOT: Final = 'not'
CONTAINS: Final = 'contains'
TARGET: Final = 'target'
THEN: Final = 'then'
OTHERWISE: Final = 'otherwise'

# Slicing
SLICING: Final = 'slicing'
FROM: Final = 'from'
TO: Final = 'to'

# Casting
# 'to' from slicing
INTEGER: Final = 'integer'
DECIMAL: Final = 'decimal'
DATE: Final = 'date'
ORIGINAL_FORMAT: Final = 'original_format'
INTEGER_CONTAINING_DECIMALS = 'integer_containing_decimals'
YMD_DATE_FORMAT: Final = r'(^(yy|yyyy)[^\w]?mm[^\w]?dd$)'
DMY_DATE_FORMAT: Final = r'(^dd[^\w]?mm[^\w]?(yy|yyyy)$)'
MDY_DATE_FORMAT = r'(^mm[^\w]?dd[^\w]?(yy|yyyy)$)'

# Casting helpers
COMMA: Final[str] = ','
PERIOD: Final[str] = '.'
EMPTY: Final[str] = ''

# ISO
ALPHA_TWO = 'alpha_2'
ALPHA_THREE = 'alpha_3'
NUMERIC = 'numeric'
OFFICIAL_NAME = 'official_name'
INVALID = 'invalid'
# NAME



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['piri2']

package_data = \
{'': ['*']}

install_requires = \
['jsonschema>=3.2.0,<4.0.0',
 'pycountry==19.8.18',
 'returns>=0.14.0,<0.15.0',
 'typing_extensions==3.7.4']

setup_kwargs = {
    'name': 'piri2',
    'version': '0.0.2',
    'description': 'Configurable and documentable Json transformation and mapping',
    'long_description': '# Piri\nConfigurable Data Mapping for mortals\n___\n![test](https://github.com/greenbird/piri/workflows/test/badge.svg)\n[![codecov](https://codecov.io/gh/greenbird/piri/branch/master/graph/badge.svg)](https://codecov.io/gh/greenbird/piri)\n[![Python Version](https://img.shields.io/pypi/pyversions/piri.svg)](https://pypi.org/project/piri/)\n[![wemake-python-styleguide](https://img.shields.io/badge/style-wemake-000000.svg)](https://github.com/wemake-services/wemake-python-styleguide)\n___\n\n**Documentation\n([Stable](https://piri.readthedocs.io/) |\n[Latest](https://piri.readthedocs.io/en/latest/)) |\n[Source Code](https://github.com/greenbird/piri) |\n[Task Tracker](https://github.com/greenbird/piri/issues)**\n\n## What is Piri\n\nPiri is a JSON to JSON mapper. That means that we read input JSON and create output JSON. How the output is created is based on instructions from a configuration file. The configuration file governs the the output structure and tells piri where in the input to find data and where to place it in the output. In addition to this Piri supports data transformation with `data casting`, `if conditions`, `combination of data from multiple places` and of course setting `default` values.\n\n__This enables you to change any input into the output you desire.__\n\n## Typical usecases\n\n* You `GET` data from api, but need to transform it for your backend system\n* `POST`ing data to an api that needs data on a different format than what your system produces\n* All your backends speak different language? pipe it through __Piri__\n* Customer delivers weirdly formatted data? Use __Piri__ to make it sexy\n* Have CSV but need nicely structured JSON? make CSV into a JSON list and transform it with __Piri__\n* Have XML but need to change it? make it into JSON, transform it with __Piri__ and then dump it to XML again.\n* Customers legacy system needs CSV. Use __Piri__ to transform your nicely structured JSON data into a JSON List that can be easily dumped to CSV\n\n## Official Open Piri Solutions\n\n[piri-cli](https://github.com/greenbird/piri-cli), commandline interface for file to file mapping.\n\n[piri-web](https://github.com/greenbird/piri-web), One Click deploy Web REST API for Piri JSON mapping.\n\n## Enterprise solutions\n\nComing...\n\n## Goal\n\nThe goal of this library is to make JSON to JSON transformation/mapping easy, configurable and documentable. We achieve this by using a simple but feature-rich JSON configuration which then also acts as documentation and as a contract between parties.\n\n## Why\n\nPiri was born because we really dislike mapping. Documenting whatever decisions made in your code so that some product owner understands it is also _no me gusto_. Transforming data from one format to another is something software engineers do allmost daily... It should be easy! And documenting it shouldn\'t be something you have to worry about.\n\nAfter the Worst POC in History we never wanted to do mapping by scripts and code again. This lead to the idea that it should be possible to create a file which governs how the structure should look and how the data should be transformed. This would then be the `single source of truth` and with Piri we have achieved this.\n\nWe believe that this will make collaboration between teams faster and easier. Use Piri to agree with data formats between Front-end and Back-end. Between the 3rd party system and your back-end. You can even use Piri for testing existing integrations ;-)\n\n## Features\n\n* Mapping with configuration File.\n* [JSON Schema](https://json-schema.org/) validation of the config file.\n* Structurally Transform JSON\n* Combine multiple values to one.\n* Default values\n* If statements\n    * is, contains, not\n* casting\n    * integer, decimal, iso date\n\n## Contributing\nPlease see [contribute](https://piri.readthedocs.io/en/stable/contributing)\n\n## Installation\n\nPackage is on pypi. Use pip or poetry to install\n\n```sh\npip install piri\n```\n```sh\npoetry add piri\n```\n\n## Introduction\n\nHave a look at our introduction course [here](https://piri.readthedocs.io/en/stable/introduction)\n\n## Quickstart\n```python\nimport simplejson\n\nfrom piri.process import process\n\nmy_config = {\n    \'name\': \'schema\',\n    \'array\': False,\n    \'objects\': [\n        {\n            \'name\': \'invoices\',\n            \'array\': True,\n            \'iterables\': [\n                {\n                    \'alias\': \'invoices\',\n                    \'path\': [\'root\', \'invoices\'],\n                },\n            ],\n            \'attributes\': [\n                {\n                    \'name\': \'amount\',\n                    \'mappings\': [\n                        {\n                            \'path\': [\'invoices\', \'amount\'],\n                        },\n                    ],\n                    \'casting\': {\n                        \'to\': \'decimal\',\n                        \'original_format\': \'integer_containing_decimals\',\n                    },\n                    \'default\': 0,\n                },\n                {\n                    \'name\': \'debtor\',\n                    \'mappings\': [\n                        {\n                            \'path\': [\'root\', \'customer\', \'first_name\'],\n                        },\n                        {\n                            \'path\': [\'root\', \'customer\', \'last_name\'],\n                        },\n                    ],\n                    \'separator\': \' \',\n                },\n            ],\n            \'objects\': [],\n        },\n    ],\n}\n\nexample_data = {\n    \'root\': {\n        \'customer\': {\n            \'first_name\': \'John\',\n            \'last_name\': \'Smith\',\n        },\n        \'invoices\': [\n            {\n                \'amount\': 10050,\n            },\n            {\n                \'amount\': 20050,\n            },\n            {\n                \'amount\': -15005,\n            },\n        ],\n    },\n}\n\nmapped_data = process(example_data, my_config)\n\nwith open(\'resultfile.json\', \'w\') as output_file:\n    output_file.write(simplejson.dumps(mapped_data))\n\n```\n\ncontents of resultfile.json\n```json\n{\n    "invoices": [\n        {\n            "amount": 100.5,\n            "debtor": "John Smith"\n        },\n        {\n            "amount": 200.5,\n            "debtor": "John Smith"\n        },\n        {\n            "amount": -150.05,\n            "debtor": "John Smith"\n        }\n    ]\n}\n```\n',
    'author': 'Thomas Borgen',
    'author_email': 'thomas.borgen@greenbird.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/greenbird/piri',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
